import copy
from copy import copy
from typing import Callable, Dict, List

import numpy as np
from joblib import Parallel, delayed

from . import acquisition_fun as AcquisitionFunction
from .base import BaseBO
from .solution import Solution

__authors__ = ["Hao Wang"]


class BO(BaseBO):
    """The sequential Bayesian Optimization class"""

    def _create_acquisition(self, fun: Callable = None, par: Dict = None, **kwargv):
        fun = fun if fun is not None else self._acquisition_fun
        par = {} if par is None else par
        if hasattr(getattr(AcquisitionFunction, fun), "plugin"):
            if "plugin" not in par:
                par.update({"plugin": self.fmin})

        return super()._create_acquisition(fun, par, **kwargv)

    def pre_eval_check(self, X: List) -> List:
        """Check for the duplicated solutions as it is not allowed in noiseless cases"""
        if not isinstance(X, Solution):
            X = Solution(X, var_name=self.var_names, n_obj=self.n_obj)

        N = X.N
        if hasattr(self, "data"):
            X = X + self.data

        _ = []
        for i in range(N):
            x = X[i]
            idx = np.arange(len(X)) != i
            CON = np.all(
                np.isclose(
                    np.asarray(X[idx][:, self.r_index], dtype="float"),
                    np.asarray(x[self.r_index], dtype="float"),
                ),
                axis=1,
            )
            INT = np.all(X[idx][:, self.i_index] == x[self.i_index], axis=1)
            CAT = np.all(X[idx][:, self.d_index] == x[self.d_index], axis=1)
            if not any(CON & INT & CAT):
                _ += [i]

        return X[_].tolist()


class ParallelBO(BO):
    """Batch-sequential Bayesian Optimization, which proposes multiple points in each iteration

    This class implements the multi-acquisition function approach, which samples multiple
    hyperparameter values for the acquisition function.

    """

    def __init__(
        self,
        n_point: int = 3,
        acquisition_fun: str = "MGFI",
        acquisition_par: Dict = {"t": 2},
        **kwargs,
    ):
        super().__init__(
            n_point=n_point,
            acquisition_fun=acquisition_fun,
            acquisition_par=acquisition_par,
            **kwargs,
        )
        assert self.n_point > 1

        if self._acquisition_fun == "MGFI":
            self._par_name = "t"
            # Log-normal distribution for `t` supported on [0, \infty)
            self._sampler = lambda x: np.exp(np.log(x["t"]) + 0.5 * np.random.randn())
        elif self._acquisition_fun == "UCB":
            self._par_name = "alpha"
            # Logit-normal distribution for `alpha` supported on [0, 1]
            self._sampler = lambda x: 1 / (
                1 + np.exp((x["alpha"] * 4 - 2) + 0.6 * np.random.randn())
            )
        elif self._acquisition_fun == "EpsilonPI":
            self._par_name = "epsilon"
            self._sampler = None  # TODO: implement this!
        else:
            raise NotImplementedError

        _criterion = getattr(AcquisitionFunction, self._acquisition_fun)()
        if self._par_name not in self._acquisition_par:
            self._acquisition_par[self._par_name] = getattr(_criterion, self._par_name)

    def _batch_arg_max_acquisition(self, n_point: int, return_dx: bool, fixed: Dict = None):
        criteria = []
        for _ in range(n_point):
            _par = self._sampler(self._acquisition_par)
            _acquisition_par = copy(self._acquisition_par)
            _acquisition_par.update({self._par_name: _par})
            criteria.append(
                self._create_acquisition(par=_acquisition_par, return_dx=return_dx, fixed=fixed)
            )

        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self.logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self.logger)) for _ in criteria]

        return tuple(zip(*__))


class AnnealingBO(ParallelBO):
    def __init__(self, t0: float = 2, tf: float = 1e-1, schedule: str = "exp", *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.t0 = t0
        self.tf = tf
        self.schedule = schedule
        self._acquisition_par["t"] = t0

        # TODO: add supports for UCB and epsilon-PI
        max_iter = (self.max_FEs - self._DoE_size) / self.n_point
        if self.schedule == "exp":  # exponential
            self.alpha = (self.tf / t0) ** (1.0 / max_iter)
            self._anealer = lambda t: t * self.alpha
        elif self.schedule == "linear":
            self.eta = (t0 - self.tf) / max_iter  # linear
            self._anealer = lambda t: t - self.eta
        elif self.schedule == "log":
            self.c = self.tf * np.log(max_iter + 1)  # logarithmic
            self._anealer = lambda t: t * self.c / np.log(self.iter_count + 2)
        else:
            raise NotImplementedError

        def callback():
            self._acquisition_par["t"] = self._anealer(self._acquisition_par["t"])

        self._acquisition_callbacks += [callback]


# TODO: write test file for this class
class SelfAdaptiveBO(ParallelBO):
    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        assert self.n_point > 1

    def _batch_arg_max_acquisition(self, n_point: int, return_dx: bool, fixed: Dict = None):
        criteria = []
        _t_list = []
        N = int(n_point / 2)

        for _ in range(n_point):
            _t = np.exp(self._acquisition_par["t"] * np.random.randn())
            _t_list.append(_t)
            _acquisition_par = copy(self._acquisition_par)
            _acquisition_par.update({"t": _t})
            criteria.append(
                self._create_acquisition(par=_acquisition_par, return_dx=return_dx, fixed=fixed)
            )

        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self.logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self.logger)) for _ in criteria]

        # NOTE: this adaptation is different from what I did in the LeGO paper..
        idx = np.argsort(__[1])[::-1][:N]
        self._acquisition_par["t"] = np.mean([_t_list[i] for i in idx])
        return tuple(zip(*__))


class NoisyBO(ParallelBO):
    """Bayesian Optimization for Noisy Scenarios"""

    def pre_eval_check(self, X: List):
        if not isinstance(X, Solution):
            X = Solution(X, var_name=self.var_names)
        return X

    def _create_acquisition(self, fun: Callable = None, par: Dict = None, **kwargv):
        par = {} if par is None else par
        if hasattr(getattr(AcquisitionFunction, self._acquisition_fun), "plugin"):
            # use the model prediction to determine the plugin under noisy scenarios
            # TODO: add more options for determining the plugin value
            y_ = self.model.predict(self.data)
            plugin = np.min(y_) if self.minimize else np.max(y_)
            par.update({"plugin": plugin})

        return super()._create_acquisition(fun, par, **kwargv)



import functools
from typing import Dict, List, Tuple, Union

import numpy as np
from joblib import Parallel, delayed
from sklearn.metrics import mean_absolute_percentage_error, r2_score
from sklearn.preprocessing import StandardScaler
from torch.tensor import Tensor

from .acquisition_optim import argmax_restart
from .bayes_opt import BO
from .extra.multi_objective import Hypervolume, is_non_dominated
from .extra.multi_objective.box_decompositions import NondominatedPartitioning
from .multi_objective import EHVI
from .solution import Solution
from .utils import dynamic_penalty, partial_argument, timeit

__authors__ = ["Hao Wang"]


class BaseMOBO(BO):
    """Base class for multi-objective BO"""

    def __init__(self, n_obj: int = 2, minimize: Union[bool, List[bool]] = True, **kwargv):
        # NOTE: setting `minimize=True` to make parent's constructor work
        super().__init__(minimize=True, **kwargv)
        self._check_obj_fun(n_obj)
        self._check_minimize(minimize)

    def _check_minimize(self, minimize):
        self.minimize = minimize
        if isinstance(self.minimize, bool):
            self.minimize = [self.minimize] * self.n_obj
        elif hasattr(self.minimize, "__iter__"):
            assert len(self.minimize) == self.n_obj
        self.minimize = np.asarray(self.minimize)

    def _check_obj_fun(self, n_obj):
        """check the objective functions"""
        self.n_obj = n_obj
        if self.obj_fun is None:
            return
        assert hasattr(self.obj_fun, "__iter__")
        assert all([hasattr(f, "__call__") for f in self.obj_fun])
        if len(self.obj_fun) != n_obj:
            self.logger.warning(
                f"len(obj_fun) ({self.obj_fun}) != n_obj ({n_obj})."
                "Setting n_obj according to the former."
            )
            self.n_obj = len(self.obj_fun)
        assert self.n_obj > 1

    @property
    def xopt(self):
        """get the non-dominated subset of solutions"""
        if self.y is None:
            return None
        idx = is_non_dominated(Tensor(self.y)).detach().cpu().numpy()
        return self.data[idx, :]

    @property
    def ref_point(self):
        """reference point for computing the hypervolumne"""
        # NOTE: assume maximization
        return np.min(self.y, axis=0) * 0.8

    @property
    def y(self):
        """The transformed objective value for the internal use"""
        # TODO: double check if this is still needed for GPRs
        # if any(np.isclose(self.frange, 0)) and len(y) > 1:
        # raise Exception("flat objective value!")
        if not hasattr(self, "data"):
            return None

        # Standardization should make the GP prior less mis-specified
        # self._scaler = StandardScaler()
        # y = self._scaler.fit_transform(self.data.fitness)
        y = self.data.fitness
        # convert to maximization problem
        return y * (-1) ** self.minimize

    @timeit
    def evaluate(self, X) -> List[Tuple[float]]:
        """evaluation

        Returns
        -------
        List[Tuple[float]]
            of the shape [[f1(x1),...,fm(x1)], [f1(x2),...,fm(x2)],...,[f1(xn),...,fm(xn)]]
        """
        # TODO: fix for `self.parallel_obj_fun`
        func_vals = []
        for f in self.obj_fun:
            if self.n_job > 1:  # or by ourselves..
                func_vals.append(Parallel(n_jobs=self.n_job)(delayed(f)(x) for x in X))
            else:  # or sequential execution
                func_vals.append([f(x) for x in X])

        return list(zip(*func_vals))

    @timeit
    def recommend(self) -> Union[None, List[Solution]]:
        return self.xopt

    def tell(
        self,
        X: List[Union[list, dict]],
        func_vals: List[List[float]],
        h_vals: List[List[float]] = None,
        g_vals: List[List[float]] = None,
        index: List[str] = None,
        warm_start: bool = False,
    ):
        """Tell the BO about the function values of proposed candidate solutions

        Parameters
        ----------
        X : List of Lists or Solution
            The candidate solutions which are usually proposed by the `self.ask` function
        func_vals : List/np.ndarray of float
            The corresponding function values
        """
        # TODO: implement method to handle known, expensive constraints `h_vals` and `g_vals`
        # add extra columns h_vals, g_vals to the `Solution` object
        X = self._to_geno(X, index)
        self.logger.info(f"observing {len(X)} points:")

        # if h_vals is not None or g_vals is not None:
        # raise NotImplementedError("will be implemented soon :)")

        for i, _ in enumerate(X):
            X[i].n_eval += 1
            for k in range(self.n_obj):
                X[i].fitness[k] = func_vals[i][k]
                if not warm_start:
                    self.eval_count += 1
            _fitness = ", ".join([f"f{_ + 1}: {func_vals[i][_]}" for _ in range(self.n_obj)])
            self.logger.info(f"#{i + 1} - {_fitness}, solution: {self._to_pheno(X[i])}")

        X = self.post_eval_check(X)
        if self.data_file is not None:
            X.to_csv(self.data_file, header=False, append=True)
        self.data = self.data + X if hasattr(self, "data") else X

        # update the surrogate models
        self.update_model()
        # compute the hypervolume indicator value of the incumbent
        xopt = self.xopt
        pf = self.xopt.fitness * (-1) ** self.minimize
        hv = Hypervolume(ref_point=Tensor(self.ref_point))
        self.logger.info(f"Efficient set/Pareto front (xopt/fopt):\n{xopt}")
        self.logger.info(f"Hypervolume indicator value: {hv.compute(Tensor(pf))}")

        if self.h is not None or self.g is not None:
            penalty = np.array(
                [dynamic_penalty(x, 1, self._h, self._g, minimize=False) for x in xopt]
            ).ravel()
            self.logger.info(f"... with corresponding penalty: {penalty}")

        if not warm_start:
            self.iter_count += 1
            self.hist_f.append(xopt.fitness)

    def update_model(self):
        # TODO: add model selection
        X, y = self.data, self.y
        self.model.fit(X, y)
        y_ = self.model.predict(X)
        r2 = r2_score(y, y_, multioutput="raw_values")
        MAPE = mean_absolute_percentage_error(y, y_, multioutput="raw_values")

        for i in range(self.n_obj):
            _r2 = r2[i] if not isinstance(r2, float) else r2
            _mape = MAPE[i] if not isinstance(MAPE, float) else MAPE
            self.logger.info(f"model of f{i + 1} r2: {_r2}, MAPE: {_mape}")


class MOBO(BaseMOBO):
    """EHVI with Kriging believer"""

    def __init__(self, *args, **kwargv):
        super().__init__(*args, **kwargv)
        self.AQ_max_FEs = 1e3 * self.search_space.dim
        if self._acquisition_fun != "EHVI":
            self._acquisition_fun = "EHVI"
            self.logger.warning(
                "MOBO only allows using `EHIV` acquisition function. Ignore user's argument."
            )

    def _create_acquisition(self, fixed: Dict = None, **kwargv):
        # TODO: implement the Kriging believer strategy
        partitioning = NondominatedPartitioning(ref_point=Tensor(self.ref_point), Y=Tensor(self.y))
        criterion = EHVI(
            model=self.model, ref_point=self.ref_point.tolist(), partitioning=partitioning
        )
        return partial_argument(
            functools.partial(criterion),
            self.search_space.var_name,
            fixed,
            reduce_output=False,
        )

    # def _batch_arg_max_acquisition(
    #     self, n_point: int, return_dx: bool, fixed: Dict = None
    # ) -> Tuple[List[list], List[float]]:
    #     """Set ``self._argmax_restart`` for optimizing the acquisition function"""
    #     candidates, values = self._argmax_restart(
    #         self._create_acquisition(par={"n_point": n_point, "model": self.model}, fixed=fixed),
    #         logger=self.logger,
    #     )
    #     return [candidates], [values]


class MOBO_qEHVI(BaseMOBO):
    """qEHVI (q-point Expected Hypervolume Improvement)"""

    def __init__(self, *args, **kwargv):
        super().__init__(*args, **kwargv)
        if self._acquisition_fun != "qEHVI":
            self._acquisition_fun = "qEHVI"
            self.logger.warning(
                "MOBO only allows using `qEHIV` acquisition function. Ignore user's argument."
            )

    # def _create_acquisition(
    #     self, fun: str = None, par: dict = None, return_dx: bool = False, fixed: Dict = None
    # ):
    #     ref_point = np.max(self.y, axis=0) * 1.3
    #     partitioning = NondominatedPartitioning(ref_point=Tensor(ref_point), Y=Tensor(self.y))
    #     return qEHVI(
    #         model=par["model"],
    #         ref_point=ref_point.tolist(),
    #         n_point=par["n_point"],
    #         partitioning=partitioning,
    #     )

    def _batch_arg_max_acquisition(
        self, n_point: int, return_dx: bool, fixed: Dict = None
    ) -> Tuple[List[list], List[float]]:
        """Set ``self._argmax_restart`` for optimizing the acquisition function"""
        fixed = {} if fixed is None else fixed
        mask = np.array([v in fixed.keys() for v in self._search_space.var_name])
        values = [fixed[k] for i, k in enumerate(self._search_space.var_name) if mask[i]]
        idx = np.nonzero(~mask)[0]
        _argmax_restart = functools.partial(
            argmax_restart,
            search_space=self._search_space[idx] * n_point,
            h=((self._h, np.repeat(mask, n_point), values * n_point) if self._h else None),
            g=(
                partial_argument(self._g, np.repeat(mask, n_point), values * n_point)
                if self._g
                else None
            ),
            eval_budget=self.AQ_max_FEs,
            n_restart=self.AQ_n_restart,
            wait_iter=self.AQ_wait_iter,
            optimizer=self._optimizer,
        )
        candidates, values = _argmax_restart(
            self._create_acquisition(par={"n_point": n_point, "model": self.model}, fixed=fixed),
            logger=self.logger,
        )
        candidates = [candidates[i * self.dim : (i + 1) * self.dim] for i in range(n_point)]
        return candidates, [values] * n_point



import functools
from copy import copy
from typing import Dict, Union

import numpy as np
from joblib import Parallel, delayed
from scipy.stats import norm, rankdata
from sklearn.decomposition import PCA, KernelPCA

from . import acquisition_fun as AcquisitionFunction
from .acquisition_fun import penalized_acquisition
from .bayes_opt import ParallelBO
from .search_space import RealSpace


class PCABO(ParallelBO):
    """Dimensionality reduction using Principle Component Decomposition (PCA)"""

    def __init__(self, kernel_pca: bool = False, n_components: Union[float, int] = None, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(self._search_space, RealSpace)

        self.__search_space = self._search_space  # the original search space
        self.kernel_pca = kernel_pca
        self._n_components = n_components

    def _scale_X(self, X, func_vals):
        self._X_mean = X.mean(axis=0)
        X_ = X - self._X_mean

        if not self.minimize:
            func_vals = -1 * func_vals

        r = rankdata(func_vals)
        N = len(func_vals)
        w = np.log(N) - np.log(r)
        w /= np.sum(w)
        return X_ * w.reshape(-1, 1)

    def _compute_bounds(self, pca, search_space):
        C = np.array([(l + u) / 2 for l, u in search_space.bounds])
        radius = norm(np.np.array([l for l, _ in search_space.bounds]) - C)
        C = C - pca.mean_ - self._X_mean
        C_ = C.dot(pca.components_.T)
        return [(_ - radius, _ + radius) for _ in C_]

    def _create_acquisition(self, fun=None, par={}, return_dx=False):
        acquisition_func = super()._create_acquisition(fun, par, return_dx)
        fun = functools.partial(
            penalized_acquisition,
            acquisition_func=acquisition_func,
            X_mean=self._X_mean,
            pca=self._pca,
            bounds=self.__search_space.bounds,
            return_dx=return_dx,
        )
        return fun

    def ask(self, n_point=None):
        X = super().ask(n_point)
        if hasattr(self, "_pca"):
            X = self._pca.inverse_transform(X) + self._X_mean
        return X

    def tell(self, X, func_vals):
        X_ = self._scale_X(X, func_vals)

        if self.kernel_pca:
            # TODO: finish the kernel PCA part..
            self._pca = KernelPCA(kernel="rbf", fit_inverse_transform=True, gamma=10)
        else:
            self._pca = PCA(n_components=self._n_components, svd_solver="full")

        X_ = self._pca.fit_transform(X_, func_vals)
        bounds = self._compute_bounds(self._pca, self.__search_space)

        # set the search space in the reduced (feature) space
        self._search_space = RealSpace(bounds)
        super().tell(X_, func_vals)


class MultiAcquisitionBO(ParallelBO):
    """Using multiple acquisition functions for parallelization"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.n_point > 1

        self._acquisition_fun = "MGFI"
        self._acquisition_fun_list = ["MGFI", "UCB"]
        self._sampler_list = [
            lambda x: np.exp(np.log(x["t"]) + 0.5 * np.random.randn()),
            lambda x: 1 / (1 + np.exp((x["alpha"] * 4 - 2) + 0.6 * np.random.randn())),
        ]
        self._par_name_list = ["t", "alpha"]
        self._acquisition_par_list = [{"t": 1}, {"alpha": 0.1}]
        self._N_acquisition = len(self._acquisition_fun_list)

        for i, _n in enumerate(self._par_name_list):
            _criterion = getattr(AcquisitionFunction, self._acquisition_fun_list[i])()
            if _n not in self._acquisition_par_list[i]:
                self._acquisition_par_list[i][_n] = getattr(_criterion, _n)

    def _batch_arg_max_acquisition(self, n_point: int, return_dx: bool, fixed: Dict = None):
        criteria = []
        for i in range(n_point):
            k = i % self._N_acquisition
            _acquisition_fun = self._acquisition_fun_list[k]
            _acquisition_par = self._acquisition_par_list[k]
            _par = self._sampler_list[k](_acquisition_par)
            _acquisition_par = copy(_acquisition_par)
            _acquisition_par.update({self._par_name_list[k]: _par})
            criteria.append(
                self._create_acquisition(
                    fun=_acquisition_fun, par=_acquisition_par, return_dx=return_dx, fixed=fixed
                )
            )

        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self.logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self.logger)) for _ in criteria]

        return tuple(zip(*__))



import logging
from typing import Any, Callable, List, Tuple, Union

import numpy as np
from scipy.optimize import fmin_l_bfgs_b

from ..search_space import RealSpace
from .mies import MIES
from .one_plus_one_cma_es import OnePlusOne_Cholesky_CMA, OnePlusOne_CMA

__all__ = ["argmax_restart", "OnePlusOne_CMA", "OnePlusOne_Cholesky_CMA", "MIES"]


def argmax_restart(
    obj_func: Callable,
    search_space,
    h: Callable = None,
    g: Callable = None,
    eval_budget: int = 100,
    n_restart: int = 10,
    wait_iter: int = 3,
    optimizer: str = "BFGS",
    logger: logging.Logger = None,
):
    # lists of the best solutions and acquisition values from each restart
    xopt, fopt = [], []
    best = -np.inf
    wait_count = 0
    if not isinstance(search_space, RealSpace) and optimizer == "BFGS":
        optimizer = "MIES"
        logger.warning("L-BFGS-B cannot be applied on continuous search space")

    if (h is not None or g is not None) and optimizer == "BFGS":
        optimizer = "OnePlusOne_Cholesky_CMA"
        # TODO: add constraint handling for BFGS
        logger.warning("L-BFGS-B cannot be applied with constraints at this moment")

    for iteration in range(n_restart):
        x0 = search_space.sample(N=1, method="uniform")[0]

        if optimizer == "BFGS":
            bounds = np.array(search_space.bounds)

            if not all([isinstance(_, float) for _ in x0]):
                raise ValueError("BFGS is not supported with mixed variable types.")

            func = lambda x: tuple(map(lambda x: -1.0 * x, obj_func(x)))
            xopt_, fopt_, stop_dict = fmin_l_bfgs_b(
                func, x0, pgtol=1e-8, factr=1e6, bounds=bounds, maxfun=eval_budget
            )

            xopt_ = xopt_.flatten().tolist()
            if not isinstance(fopt_, float):
                fopt_ = float(fopt_)
            fopt_ = -fopt_

            if logger is not None and stop_dict["warnflag"] != 0:
                logger.debug("L-BFGS-B terminated abnormally with the state: %s" % stop_dict)
        elif optimizer == "OnePlusOne_Cholesky_CMA":
            lb, ub = list(zip(*search_space.bounds))
            opt = OnePlusOne_Cholesky_CMA(
                dim=search_space.dim,
                obj_fun=obj_func,
                h=h,
                g=g,
                lb=lb,
                ub=ub,
                max_FEs=eval_budget,
                ftol=1e-4,
                xtol=1e-4,
                n_restart=0,
                minimize=False,
                verbose=False,
            )
            xopt_, fopt_, stop_dict = opt.run()
            stop_dict["funcalls"] = stop_dict["FEs"] if "FEs" in stop_dict else opt.eval_count

        elif optimizer == "MIES":
            xopt_, fopt_, stop_dict = MIES(
                search_space,
                obj_func,
                eq_func=h,
                ineq_func=g,
                max_eval=eval_budget,
                minimize=False,
                verbose=False,
                eval_type="list",
            ).optimize()

        if fopt_ > best:
            best = fopt_
            wait_count = 0

            if logger is not None:
                logger.debug(
                    "restart : %d - funcalls : %d - Fopt : %f"
                    % (iteration + 1, stop_dict["funcalls"], fopt_)
                )
        else:
            wait_count += 1

        eval_budget -= stop_dict["funcalls"]
        xopt.append(xopt_)
        fopt.append(fopt_)

        if eval_budget <= 0 or wait_count >= wait_iter:
            break

    idx = np.argsort(fopt)[::-1]
    return xopt[idx[0]], fopt[idx[0]]



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="bayes-optim",
    version="0.2.1",
    author="Hao Wang",
    author_email="wangronin@gmail.com",
    packages=find_packages(),
    description="A Bayesian Optimization Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wangronin/Bayesian-Optimization",
    package_dir={"bayes_optim": "bayes_optim"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    download_url="https://github.com/wangronin/Bayesian-Optimization/archive/v0.1.3.tar.gz",
    python_requires=">=3.6",
    install_requires=[
        "dill>=0.3.2",
        "joblib>=0.16.0",
        "numpy>=1.19.0",
        "pyDOE==0.3.8",
        "scikit-learn>=0.23.0",
        "scipy>=1.5.0",
        "sklearn==0.0",
        "tabulate>=0.8.7",
        "threadpoolctl>=2.1.0",
        "torch==1.7.1",
    ],
)



from __future__ import annotations

import torch
from torch import Tensor


def is_non_dominated(Y: Tensor, deduplicate: bool = True) -> Tensor:
    r"""Computes the non-dominated front.

    Note: this assumes maximization.

    Args:
        Y: A `(batch_shape) x n x m`-dim tensor of outcomes.
        deduplicate: A boolean indicating whether to only return
            unique points on the pareto frontier.

    Returns:
        A `(batch_shape) x n`-dim boolean tensor indicating whether
        each point is non-dominated.
    """
    if Y.shape[-2] == 0:
        return torch.zeros(Y.shape[:-1], dtype=torch.bool, device=Y.device)
    Y1 = Y.unsqueeze(-3)
    Y2 = Y.unsqueeze(-2)
    dominates = (Y1 >= Y2).all(dim=-1) & (Y1 > Y2).any(dim=-1)
    nd_mask = ~(dominates.any(dim=-1))
    if deduplicate:
        # remove duplicates
        # find index of first occurrence  of each unique element
        indices = (Y1 == Y2).all(dim=-1).long().argmax(dim=-1)
        keep = torch.zeros_like(nd_mask)
        keep.scatter_(dim=-1, index=indices, value=1.0)
        return nd_mask & keep
    return nd_mask
